@@ -1,6 +1,6 @@
 // Code to load disk image and start system boot.
 //
-// Copyright (C) 2008  Kevin O'Connor <kevin@koconnor.net>
+// Copyright (C) 2008-2010  Kevin O'Connor <kevin@koconnor.net>
 // Copyright (C) 2002  MandrakeSoft S.A.
 //
 // This file may be distributed under the terms of the GNU LGPLv3 license.
@@ -29,9 +29,9 @@ boot_setup(void)
     dprintf(3, "init boot device ordering\n");
 
     memset(&IPL, 0, sizeof(IPL));
+    struct ipl_entry_s *ie = &IPL.bev[0];
 
     // Floppy drive
-    struct ipl_entry_s *ie = &IPL.bev[0];
     ie->type = IPL_TYPE_FLOPPY;
     ie->description = "Floppy";
     ie++;
@@ -44,11 +44,11 @@ boot_setup(void)
     // CDROM
     if (CONFIG_CDROM_BOOT) {
         ie->type = IPL_TYPE_CDROM;
-        ie->description = "CD-Rom";
+        ie->description = "DVD/CD";
         ie++;
     }
 
-    if (CONFIG_COREBOOT_FLASH) {
+    if (CONFIG_COREBOOT && CONFIG_COREBOOT_FLASH) {
         ie->type = IPL_TYPE_CBFS;
         ie->description = "CBFS";
         ie++;
@@ -159,9 +159,7 @@ menu_show_floppy(struct ipl_entry_s *ie, int menupos)
     int i;
     for (i = 0; i < Drives.floppycount; i++) {
         struct drive_s *drive_g = getDrive(EXTTYPE_FLOPPY, i);
-        printf("%d. Floppy [", menupos + i);
-        describe_drive(drive_g);
-        printf("]\n");
+        printf("%d. Floppy [%s]\n", menupos + i, drive_g->desc);
     }
     return Drives.floppycount;
 }
@@ -173,11 +171,10 @@ menu_show_harddisk(struct ipl_entry_s *ie, int menupos)
     int i;
     for (i = 0; i < IPL.bcvcount; i++) {
         struct ipl_entry_s *ie = &IPL.bcv[i];
+        struct drive_s *drive_g = (void*)ie->vector;
         switch (ie->type) {
         case BCV_TYPE_INTERNAL:
-            printf("%d. ", menupos + i);
-            describe_drive((void*)ie->vector);
-            printf("\n");
+            printf("%d. %s\n", menupos + i, drive_g->desc);
             break;
         default:
             menu_show_default(ie, menupos+i);
@@ -194,9 +191,7 @@ menu_show_cdrom(struct ipl_entry_s *ie, int menupos)
     int i;
     for (i = 0; i < Drives.cdcount; i++) {
         struct drive_s *drive_g = getDrive(EXTTYPE_CD, i);
-        printf("%d. CD-Rom [", menupos + i);
-        describe_drive(drive_g);
-        printf("]\n");
+        printf("%d. DVD/CD [%s]\n", menupos + i, drive_g->desc);
     }
     return Drives.cdcount;
 }
@@ -232,7 +227,9 @@ interactive_bootmenu(void)
 
     printf("Press F12 for boot menu.\n\n");
 
+    enable_bootsplash();
     int scan_code = get_keystroke(CONFIG_BOOTMENU_WAIT);
+    disable_bootsplash();
     if (scan_code != 0x86)
         /* not F12 */
         return;
@@ -241,6 +238,7 @@ interactive_bootmenu(void)
         ;
 
     printf("Select boot device:\n\n");
+    wait_threads();
 
     int subcount[ARRAY_SIZE(IPL.bev)];
     int menupos = 1;
@@ -311,17 +309,20 @@ run_bcv(struct ipl_entry_s *ie)
 void
 boot_prep(void)
 {
-    if (! CONFIG_BOOT)
+    if (! CONFIG_BOOT) {
+        wait_threads();
         return;
+    }
 
     // XXX - show available drives?
 
     // Allow user to modify BCV/IPL order.
     interactive_bootmenu();
+    wait_threads();
 
     // Setup floppy boot order
     int override = IPL.bev[0].subchoice;
-    int tmp = Drives.idmap[EXTTYPE_FLOPPY][0];
+    struct drive_s *tmp = Drives.idmap[EXTTYPE_FLOPPY][0];
     Drives.idmap[EXTTYPE_FLOPPY][0] = Drives.idmap[EXTTYPE_FLOPPY][override];
     Drives.idmap[EXTTYPE_FLOPPY][override] = tmp;
 
@@ -345,7 +346,6 @@ static void
 call_boot_entry(u16 bootseg, u16 bootip, u8 bootdrv)
 {
     dprintf(1, "Booting from %04x:%04x\n", bootseg, bootip);
-
     struct bregs br;
     memset(&br, 0, sizeof(br));
     br.flags = F_IF;
@@ -419,7 +419,7 @@ boot_cdrom(struct ipl_entry_s *ie)
 static void
 boot_cbfs(struct ipl_entry_s *ie)
 {
-    if (! CONFIG_COREBOOT_FLASH)
+    if (!CONFIG_COREBOOT || !CONFIG_COREBOOT_FLASH)
         return;
     int count = ie->subchoice;
     struct cbfs_file *file = NULL;
@@ -450,7 +450,7 @@ do_boot(u16 seq_nr)
         printf("No bootable device.\n");
         // Loop with irqs enabled - this allows ctrl+alt+delete to work.
         for (;;)
-            biosusleep(1000000);
+            wait_irq();
     }
 
     /* Do the loading, and set up vector as a far pointer to the boot
@@ -460,7 +460,7 @@ do_boot(u16 seq_nr)
     printf("Booting from %s...\n"
            , strtcpy(desc, ie->description, ARRAY_SIZE(desc)));
 
-    switch(ie->type) {
+    switch (ie->type) {
     case IPL_TYPE_FLOPPY:
         boot_disk(0x00, IPL.checkfloppysig);
         break;
